from component.messagebus import publisher
from module.snake_module_name.component.repo.snake_entity_name_repo import (
    snake_entity_name_repo
)
from module.snake_module_name.entity.snake_entity_name.model import (
    PascalEntityNameModel
)

snake_entity_name_model: PascalEntityNameModel = PascalEntityNameModel(
    snake_entity_name_repo, publisher
)



from typing import Any
from ..task.decorator import python_task
from ..runner import runner
from ..config.config import version


@python_task(
    name='version',
    description='Get Zrb version',
    runner=runner
)
async def get_version(*args: Any, **kwargs: Any) -> str:
    return version



from zrb.helper.codemod.add_key_value_to_dict import (
    add_key_value_to_dict
)


def test_add_key_value_to_dict():
    code = add_key_value_to_dict(
        code='\n'.join([
            'person = {"name": "Bob", "age": 5}',
            'alchemist = {"name": "Eldric", "age": 15}',
        ]),
        dict_name='alchemist',
        key='"job"',
        value='"alchemist"'
    )
    assert code == '\n'.join([
        'person = {"name": "Bob", "age": 5}',
        'alchemist = {"name": "Eldric", "age": 15, "job": "alchemist"}',
    ])



from pydantic import BaseModel


class AccessTokenData(BaseModel):
    user_id: str
    username: str
    expire_seconds: int


class RefreshTokenData(BaseModel):
    user_id: str
    expire_seconds: int


class TokenResponse(BaseModel):
    access_token: str
    refresh_token: str
    token_type: str



from typing import Any, Mapping
from core.messagebus.messagebus import Consumer, Publisher
from core.rpc.rpc import Server, TRPCHandler, Message, Result
import logging
import inspect
from pydantic import BaseModel


class MessagebusServer(Server):

    def __init__(
        self,
        logger: logging.Logger,
        consumer: Consumer,
        publisher: Publisher
    ):
        self.logger = logger
        self.consumer = consumer
        self.publisher = publisher

    async def start(self):
        await self.consumer.start()

    async def stop(self):
        await self.consumer.stop()

    def register(self, rpc_name):
        def wrapper(handler: TRPCHandler):
            @self.consumer.register(rpc_name)
            async def event_handler(message_dict: Mapping[str, Any]):
                message: Message = Message.from_dict(message_dict)
                args = message.args
                kwargs = message.kwargs
                reply_event = message.reply_event
                # get reply
                self.logger.info(
                    '🤙 [messagebus-rpc-server] ' +
                    f'Invoke RPC: {rpc_name} ' +
                    f', args: {args} ' +
                    f', kwargs: {kwargs} ' +
                    f', reply_event: {reply_event}'
                )
                try:
                    result = await self._run_handler(handler, *args, **kwargs)
                    if isinstance(result, BaseModel):
                        result = result.dict()
                    # publish result
                    await self.publisher.publish(
                        reply_event, Result(result=result).to_dict()
                    )
                except Exception as e:
                    self.logger.error(e, exc_info=True)
                    # publish error
                    await self.publisher.publish(
                        reply_event, Result(error=f'{e}').to_dict()
                    )
        return wrapper

    async def _run_handler(
        self, handler: TRPCHandler, *args: Any, **kwargs: Any
    ):
        if inspect.iscoroutinefunction(handler):
            return await handler(*args, **kwargs)
        return handler(*args, **kwargs)



from typing import Any, Callable, Iterable, Mapping, Optional, Union
from typeguard import typechecked
from .base_task import BaseTask, Group
from .cmd_task import CmdTask
from ..task_env.env import Env
from ..task_env.env_file import EnvFile
from ..task_input.base_input import BaseInput
from ..helper.string.double_quote import double_quote
from ..helper.docker_compose.file import read_compose_file
from ..helper.docker_compose.fetch_external_env import (
    fetch_compose_file_env_map
)

import os
import pathlib


@typechecked
class DockerComposeTask(CmdTask):
    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        inputs: Iterable[BaseInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = '',
        executable: Optional[str] = None,
        compose_file: Optional[str] = None,
        compose_cmd: str = 'up',
        compose_options: Mapping[str, str] = {},
        compose_flags: Iterable[str] = [],
        compose_args: Iterable[str] = [],
        compose_env_prefix: str = '',
        setup_cmd: Union[str, Iterable[str]] = '',
        setup_cmd_path: str = '',
        cwd: Optional[Union[str, pathlib.Path]] = None,
        upstreams: Iterable[BaseTask] = [],
        checkers: Iterable[BaseTask] = [],
        checking_interval: float = 0.1,
        retry: int = 2,
        retry_interval: float = 1,
        max_output_line: int = 1000,
        max_error_line: int = 1000,
        preexec_fn: Optional[Callable[[], Any]] = os.setsid,
        skip_execution: Union[bool, str] = False
    ):
        combined_env_files = list(env_files)
        CmdTask.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=combined_env_files,
            icon=icon,
            color=color,
            description=description,
            executable=executable,
            cwd=cwd,
            upstreams=upstreams,
            checkers=checkers,
            checking_interval=checking_interval,
            retry=retry,
            retry_interval=retry_interval,
            max_output_line=max_output_line,
            max_error_line=max_error_line,
            preexec_fn=preexec_fn,
            skip_execution=skip_execution
        )
        self.setup_cmd = setup_cmd
        self.setup_cmd_path = setup_cmd_path
        self.compose_cmd = compose_cmd
        self.compose_options = compose_options
        self.compose_flags = compose_flags
        self.compose_args = compose_args
        self.compose_env_prefix = compose_env_prefix
        self._set_compose_file(compose_file)
        self._fill_envs()

    def _fill_envs(self):
        data = read_compose_file(self.compose_file)
        env_map = fetch_compose_file_env_map(data)
        for key, value in env_map.items():
            # Need to get this everytime because we only want
            # the first compose file env value for a certain key
            existing_env_map = self._get_existing_env_map()
            if key in existing_env_map:
                continue
            os_name = key
            if self.compose_env_prefix != '':
                os_name = f'{self.compose_env_prefix}_{os_name}'
            self.envs.append(Env(name=key, os_name=os_name, default=value))

    def _get_existing_env_map(self) -> Mapping[str, str]:
        env_map: Mapping[str, str] = {}
        for env_file in self.env_files:
            envs = env_file.get_envs()
            env_map.update({
                env.name: env.default for env in envs
            })
        env_map.update({
            env.name: env.default for env in self.envs
        })
        return env_map

    def _set_compose_file(self, compose_file: Optional[str]):
        if compose_file is None:
            for _compose_file in [
                'compose.yml', 'compose.yaml',
                'docker-compose.yml', 'docker-compose.yaml'
            ]:
                if os.path.exists(os.path.join(self.cwd, _compose_file)):
                    self.compose_file: str = os.path.join(
                        self.cwd, _compose_file
                    )
                    return
            raise Exception(f'Cannot find compose file on {self.cwd}')
        if os.path.isabs(compose_file) and os.path.exists(compose_file):
            self.compose_file: str = compose_file
            return
        if os.path.exists(os.path.join(self.cwd, compose_file)):
            self.compose_file: str = os.path.join(self.cwd, compose_file)
            return
        raise Exception(f'Invalid compose file: {compose_file}')

    def _get_cmd_str(self) -> str:
        setup_cmd_str = self._create_cmd_str(
            self.setup_cmd_path, self.setup_cmd
        )
        command_options = dict(self.compose_options)
        if '--file' not in command_options and '-f' not in command_options:
            command_options['--file'] = self.compose_file
        options = ' '.join([
            f'{self.render_str(key)} {double_quote(self.render_str(val))}'
            for key, val in command_options.items()
            if self.render_str(val) != ''
        ])
        flags = ' '.join([
            self.render_str(flag) for flag in self.compose_flags
            if self.render_str(flag) != ''
        ])
        args = ' '.join([
            double_quote(self.render_str(arg)) for arg in self.compose_args
            if self.render_str(arg) != ''
        ])
        cmd_str = '\n'.join([
            setup_cmd_str,
            f'docker compose {options} {self.compose_cmd} {flags} {args}',
        ])
        self.log_info(f'Command: {cmd_str}')
        return cmd_str



from typing import List
from zrb import Input, DockerComposeTask, runner
from zrb.builtin._group import project_group
from ._common import (
    RESOURCE_DIR,
    local_input, image_input, image_env
)

compose_inputs: List[Input] = [
    local_input,
    image_input,
]
compose_env_prefix = 'CONTAINER_ENV_PREFIX'

###############################################################################
# Task Definitions
###############################################################################

build_snake_app_name_image = DockerComposeTask(
    icon='🏭',
    name='build-kebab-app-name-image',
    description='Build human readable app name image',
    group=project_group,
    inputs=compose_inputs,
    envs=[image_env],
    skip_execution='{{not input.local_snake_app_name}}',
    cwd=RESOURCE_DIR,
    compose_cmd='build',
    compose_args=[
        'kebab-app-name'
    ],
    compose_env_prefix=compose_env_prefix,
)
runner.register(build_snake_app_name_image)

push_snake_app_name_image = DockerComposeTask(
    icon='📰',
    name='push-kebab-app-name-image',
    description='Push human readable app name image',
    group=project_group,
    inputs=compose_inputs,
    envs=[image_env],
    upstreams=[build_snake_app_name_image],
    cwd=RESOURCE_DIR,
    compose_cmd='push',
    compose_args=[
        'kebab-app-name'
    ],
    compose_env_prefix=compose_env_prefix,
)
runner.register(push_snake_app_name_image)



from component.log import logger
from component.db_connection import engine
from module.log.entity.activity.repo import (
    ActivityRepo, ActivityDBRepo
)

activity_repo: ActivityRepo = ActivityDBRepo(
    logger=logger, engine=engine
)



from typing import Any, Mapping
from ruamel.yaml import YAML, CommentedMap


def read_compose_file(file_name: str) -> Any:
    yaml = YAML()
    with open(file_name, 'r') as file:
        data = yaml.load(file)
    return data


def add_services(file_name: str, new_services: Mapping[str, str]):
    yaml = YAML()
    with open(file_name, 'r') as f:
        data = yaml.load(f)
        # data = round_trip_load(f, preserve_quotes=True)
    data = CommentedMap(data)
    data['services'].update(CommentedMap(new_services))
    # Write the modified data structure back to the file
    with open(file_name, 'w') as f:
        yaml.dump(data, f)
        # round_trip_dump(data, f)



from typing import Any


def is_probably_jinja(value: Any) -> bool:
    if not isinstance(value, str):
        return False
    if '{{' in value and '}}' in value:
        return True
    if '{%' in value and '%}' in value:
        return True
    return False



from ..task.cmd_task import CmdTask
from ..runner import runner

update = CmdTask(
    name='update',
    description='Update zrb',
    cmd=[
        'pip install zrb -U',
    ],
    checking_interval=3
)
runner.register(update)



from zrb.helper.codemod.add_import_module import add_import_module


def test_import_module():
    code = add_import_module(
        code='',
        module_path='my_package.my_module'
    )
    assert code == 'import my_package.my_module\n'


def test_import_module_relative():
    is_error = False
    try:
        add_import_module(
            code='',
            module_path='.my_package.my_module',
        )
    except Exception:
        is_error = True
    assert is_error


def test_import_module_with_alias():
    code = add_import_module(
        code='',
        module_path='my_package.my_module',
        alias='alias'

    )
    assert code == 'import my_package.my_module as alias\n'


def test_import_from_module():
    code = add_import_module(
        code='',
        module_path='my_package.my_module',
        resource='resource'
    )
    assert code == 'from my_package.my_module import resource\n'


def test_import_from_module_relative():
    code = add_import_module(
        code='',
        module_path='.my_package.my_module',
        resource='resource'
    )
    assert code == 'from .my_package.my_module import resource\n'


def test_import_from_module_with_alias():
    code = add_import_module(
        code='',
        module_path='my_package.my_module',
        resource='resource',
        alias='alias'
    )
    assert code == 'from my_package.my_module import resource as alias\n'


def test_import_from_module_with_alias_on_existing_code():
    code = add_import_module(
        code='from somewhere import something\n',
        module_path='my_package.my_module',
        resource='resource',
        alias='alias'
    )
    assert code == '\n'.join([
        'from somewhere import something',
        'from my_package.my_module import resource as alias'
    ]) + '\n'



from typing import List, Mapping
from ...task.base_task import Group, BaseTask
from ...task_env.env import Env
from ..string.jinja import is_probably_jinja


def fetch_env_map_from_group(
    env_map: Mapping[str, str], group: Group
) -> Mapping[str, str]:
    for task in group.tasks:
        env_map = fetch_env_map_from_task(env_map, task)
    for sub_group in group.children:
        sub_env_map: Mapping[str, str] = fetch_env_map_from_group(
            env_map, sub_group
        )
        env_map = cascade_env_map(env_map, sub_env_map)
    return env_map


def fetch_env_map_from_task(
    env_map: Mapping[str, str], task: BaseTask
):
    task_env_map: Mapping[str, str] = {}
    for env_file in task.env_files:
        envs = env_file.get_envs()
        task_env_map = add_envs_to_env_map(task_env_map, envs)
    task_env_map = add_envs_to_env_map(task_env_map, task.envs)
    env_map = cascade_env_map(env_map, task_env_map)
    for upstream in task.upstreams:
        task_env_map = fetch_env_map_from_task(env_map, upstream)
    for checker in task.checkers:
        task_env_map = fetch_env_map_from_task(env_map, checker)
    return env_map


def add_envs_to_env_map(
    env_map: Mapping[str, str], envs: List[Env]
) -> Mapping[str, str]:
    for env in envs:
        if env.os_name == '':
            continue
        env_name = get_env_name(env)
        env_default = get_env_default(env)
        env_map[env_name] = env_default
    return env_map


def cascade_env_map(
    env_map: Mapping[str, str],
    other_env_map: Mapping[str, str]
) -> Mapping[str, str]:
    for key, value in other_env_map.items():
        if key in env_map:
            continue
        env_map[key] = value
    return env_map


def get_env_name(env: Env) -> str:
    if env.os_name is None:
        return env.name
    return env.os_name


def get_env_default(env: Env) -> str:
    if is_probably_jinja(env.default):
        return ''
    return env.default



from typing import Any, List, Optional
from ....task.decorator import python_task
from ....task.task import Task
from ....helper.codemod.add_assert_resource import add_assert_resource
from ....helper.codemod.add_import_module import add_import_module
from ....helper.file.text import read_text_file_async, write_text_file_async
from ....helper import util
from .input import project_dir_input, task_name_input, app_name_input
import os


def validate_project_dir(project_dir: str):
    if not os.path.isfile(os.path.join(project_dir, 'zrb_init.py')):
        raise Exception(f'Not a project: {project_dir}')


def create_register_task_module(
    upstreams: Optional[List[Task]] = None
) -> Task:
    @python_task(
        name='register-task-module',
        inputs=[project_dir_input, task_name_input],
        upstreams=[] if upstreams is None else upstreams
    )
    async def task(*args: Any, **kwargs: Any):
        task: Task = kwargs.get('_task')
        project_dir = kwargs.get('project_dir')
        validate_project_dir(project_dir)
        task_name = kwargs.get('task_name')
        task.print_out(f'Register module: _automate.{task_name}')
        await register_module_to_project(
            project_dir=project_dir,
            module_name='.'.join([
                '_automate', util.to_snake_case(task_name)
            ])
        )
    return task


def create_register_app_module(
    module: str, upstreams: Optional[List[Task]] = None
) -> Task:
    @python_task(
        name=f'register-{module}',
        inputs=[project_dir_input, app_name_input],
        upstreams=[] if upstreams is None else upstreams
    )
    async def task(*args: Any, **kwargs: Any):
        task: Task = kwargs.get('_task')
        project_dir = kwargs.get('project_dir')
        validate_project_dir(project_dir)
        app_name = kwargs.get('app_name')
        task.print_out(f'Register module: _automate.{app_name}.{module}')
        snake_app_name = util.to_snake_case(app_name)
        await register_module_to_project(
            project_dir=project_dir,
            module_name='.'.join([
                '_automate', snake_app_name, module
            ]),
            alias=f'{snake_app_name}_{module}'
        )
    return task


async def register_module_to_project(
    project_dir: str, module_name: str, alias: Optional[str] = None
):
    zrb_init_path = os.path.join(project_dir, 'zrb_init.py')
    code = await read_text_file_async(zrb_init_path)
    if alias is None:
        alias = module_name.split('.')[-1]
    code = add_import_module(
        code=code,
        module_path=module_name,
        alias=alias
    )
    code = add_assert_resource(code, alias)
    await write_text_file_async(zrb_init_path, code)
    return True



from zrb.builtin.principle import (
    show_dry_principle, show_kiss_principle,
    show_solid_principle, show_yagni_principle
)


def test_show_dry_principle():
    main_loop = show_dry_principle.create_main_loop()
    result = main_loop()
    assert result is None


def test_show_kiss_principle():
    main_loop = show_kiss_principle.create_main_loop()
    result = main_loop()
    assert result is None


def test_show_solid_principle():
    main_loop = show_solid_principle.create_main_loop()
    result = main_loop()
    assert result is None


def test_show_yagni_principle():
    main_loop = show_yagni_principle.create_main_loop()
    result = main_loop()
    assert result is None



from zrb.builtin.env import (
    show, get
)


def test_show():
    main_loop = show.create_main_loop()
    result = main_loop()
    assert result is None


def test_get():
    main_loop = get.create_main_loop()
    result = main_loop()
    assert len(result.keys()) > 0



from module.auth.core.authorizer.authorizer import Authorizer
from module.auth.core.authorizer.rpc_authorizer import RPCAuthorizer
from module.auth.core.password_hasher.password_hasher import PasswordHasher
from module.auth.core.password_hasher.bcrypt_password_hasher import (
    BcryptPasswordHasher
)
from module.auth.core.access_token.scheme import (
    AccessTokenScheme, create_oauth2_bearer_access_token_scheme
)
from module.auth.core.access_token.util import (
    AccessTokenUtil, JWTAccessTokenUtil
)
from module.auth.core.refresh_token.util import (
    RefreshTokenUtil, JWTRefreshTokenUtil
)

assert Authorizer
assert RPCAuthorizer
assert PasswordHasher
assert BcryptPasswordHasher
assert AccessTokenScheme
assert create_oauth2_bearer_access_token_scheme
assert AccessTokenUtil
assert JWTAccessTokenUtil
assert RefreshTokenUtil
assert JWTRefreshTokenUtil



from sqlalchemy import Column, String
from core.repo import Repo, DBEntityMixin, DBRepo
from module.snake_module_name.schema.snake_entity_name import (
    PascalEntityName, PascalEntityNameData
)
from module.snake_module_name.component import Base


class DBEntityPascalEntityName(Base, DBEntityMixin):
    __tablename__ = "snake_plural_entity_name"
    snake_column_name = Column(String)


class PascalEntityNameRepo(Repo[PascalEntityName, PascalEntityNameData]):
    pass


class PascalEntityNameDBRepo(
    DBRepo[PascalEntityName, PascalEntityNameData], PascalEntityNameRepo
):
    schema_cls = PascalEntityName
    db_entity_cls = DBEntityPascalEntityName



from zrb.helper.codemod.add_assert_resource import add_assert_resource


def test_import_module():
    code = add_assert_resource(
        code='a = True',
        resource='a'
    )
    assert code == '\n'.join([
        'a = True',
        'assert a'
    ])



from component.messagebus import publisher
from module.auth.component.repo.group_repo import group_repo
from module.auth.entity.group.model import GroupModel

group_model: GroupModel = GroupModel(group_repo, publisher)



from sqlalchemy import Column, String, Text
from core.repo import Repo, DBEntityMixin, DBRepo
from module.log.schema.activity import (
    Activity, ActivityData
)
from module.log.component import Base


class DBEntityActivity(Base, DBEntityMixin):
    __tablename__ = "activities"
    action = Column(String)
    entity: Column = Column(String)
    data: Column = Column(Text)


class ActivityRepo(Repo[Activity, ActivityData]):
    pass


class ActivityDBRepo(
    DBRepo[Activity, ActivityData], ActivityRepo
):
    schema_cls = Activity
    db_entity_cls = DBEntityActivity



from zrb.helper.util import (
    coalesce, to_camel_case, to_pascal_case, to_kebab_case,
    to_snake_case, to_human_readable
)


def test_coalesce():
    assert coalesce('a', 'b', 'c') == 'a'
    assert coalesce(None, 'b', 'c') == 'b'
    assert coalesce(None, None, 'c') == 'c'
    assert coalesce(None, None, None) is None


def test_to_camel_case():
    assert to_camel_case('') == ''
    assert to_camel_case(' send  RPC  Request ') == 'sendRpcRequest'
    assert to_camel_case('send RPC Request') == 'sendRpcRequest'
    assert to_camel_case('sendRPCRequest') == 'sendRpcRequest'
    assert to_camel_case('SendRPCRequest') == 'sendRpcRequest'
    assert to_camel_case('send-rpc-request') == 'sendRpcRequest'
    assert to_camel_case('send_rpc_request') == 'sendRpcRequest'
    assert to_camel_case('send_rpc_request!!!') == 'sendRpcRequest'


def test_to_pascal_case():
    assert to_pascal_case('') == ''
    assert to_pascal_case(' send  RPC  Request ') == 'SendRpcRequest'
    assert to_pascal_case('send RPC Request') == 'SendRpcRequest'
    assert to_pascal_case('sendRPCRequest') == 'SendRpcRequest'
    assert to_pascal_case('SendRPCRequest') == 'SendRpcRequest'
    assert to_pascal_case('send-rpc-request') == 'SendRpcRequest'
    assert to_pascal_case('send_rpc_request') == 'SendRpcRequest'
    assert to_pascal_case('send_rpc_request!!!') == 'SendRpcRequest'


def test_to_kebab_case():
    assert to_kebab_case('') == ''
    assert to_kebab_case(' send  RPC  Request ') == 'send-rpc-request'
    assert to_kebab_case('send RPC Request') == 'send-rpc-request'
    assert to_kebab_case('sendRPCRequest') == 'send-rpc-request'
    assert to_kebab_case('SendRPCRequest') == 'send-rpc-request'
    assert to_kebab_case('send-rpc-request') == 'send-rpc-request'
    assert to_kebab_case('send_rpc_request') == 'send-rpc-request'
    assert to_kebab_case('send_rpc_request!!!') == 'send-rpc-request'


def test_to_snake_case():
    assert to_snake_case('') == ''
    assert to_snake_case(' send  RPC  Request ') == 'send_rpc_request'
    assert to_snake_case('send RPC Request') == 'send_rpc_request'
    assert to_snake_case('sendRPCRequest') == 'send_rpc_request'
    assert to_snake_case('SendRPCRequest') == 'send_rpc_request'
    assert to_snake_case('send-rpc-request') == 'send_rpc_request'
    assert to_snake_case('send_rpc_request') == 'send_rpc_request'
    assert to_snake_case('send_rpc_request!!!') == 'send_rpc_request'


def test_to_human_readable():
    assert to_human_readable('') == ''
    assert to_human_readable(' send  RPC  Request ') == 'send RPC request'
    assert to_human_readable('send RPC Request') == 'send RPC request'
    assert to_human_readable('sendRPCRequest') == 'send RPC request'
    assert to_human_readable('SendRPCRequest') == 'send RPC request'
    assert to_human_readable('send-rpc-request') == 'send rpc request'
    assert to_human_readable('send_rpc_request') == 'send rpc request'
    assert to_human_readable('send_rpc_request!!!') == 'send rpc request!!!'



from typing import Any, List, Mapping
from dotenv import dotenv_values
from ..._group import project_add_group
from ....task.task import Task
from ....task.decorator import python_task
from ....task.resource_maker import ResourceMaker
from ....runner import runner
from .._common.input import (
    project_dir_input, app_name_input,
    module_name_input
)
from .._common.helper import validate_project_dir
from .._common.lock import new_task_app_lock
from ....helper import util
from ....helper.codemod.add_import_module import add_import_module
from ....helper.codemod.add_function_call import add_function_call
from ....helper.codemod.append_code_to_function import append_code_to_function
from ....helper.docker_compose.file import add_services
from ....helper.file.text import (
    read_text_file_async, write_text_file_async, append_text_file_async
)

import asyncio
import os
import jsons

current_dir = os.path.dirname(__file__)


@python_task(
    name='validate',
    inputs=[project_dir_input, app_name_input, module_name_input],
    retry=0,
)
async def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir')
    validate_project_dir(project_dir)
    app_name = kwargs.get('app_name')
    module_name = kwargs.get('module_name')
    snake_module_name = util.to_snake_case(module_name)
    automation_dir = os.path.join(
        project_dir, '_automate', util.to_snake_case(app_name)
    )
    if not os.path.exists(automation_dir):
        raise Exception(
            f'Automation directory does not exist: {automation_dir}'
        )
    app_dir = os.path.join(
        project_dir, 'src', f'{util.to_kebab_case(app_name)}'
    )
    if not os.path.exists(app_dir):
        raise Exception(f'App directory does not exist: {app_dir}')
    module_path = os.path.join(app_dir, 'src', 'module', snake_module_name)
    if os.path.exists(module_path):
        raise Exception(f'Module directory already exists: {module_path}')


copy_resource = ResourceMaker(
    name='copy-resource',
    inputs=[
        project_dir_input,
        app_name_input,
        module_name_input,
    ],
    upstreams=[validate],
    replacements={
        'appName': '{{input.app_name}}',
        'moduleName': '{{input.module_name}}',
    },
    template_path=os.path.join(current_dir, 'template'),
    destination_path='{{ input.project_dir }}',
    locks=[new_task_app_lock],
    excludes=[]
)


@python_task(
    name='fastapp-module',
    group=project_add_group,
    upstreams=[copy_resource],
    runner=runner
)
async def add_fastapp_module(*args: Any, **kwargs: Any):
    task: Task = kwargs.get('_task')
    project_dir = kwargs.get('project_dir', '.')
    app_name = kwargs.get('app_name')
    module_name = kwargs.get('module_name')
    kebab_app_name = util.to_kebab_case(app_name)
    snake_app_name = util.to_snake_case(app_name)
    kebab_module_name = util.to_kebab_case(module_name)
    snake_module_name = util.to_snake_case(module_name)
    upper_snake_module_name = snake_module_name.upper()
    await asyncio.gather(
        asyncio.create_task(_create_microservice_config(
            task, project_dir, kebab_app_name, snake_app_name,
            kebab_module_name, snake_module_name, upper_snake_module_name
        )),
        asyncio.create_task(_register_module(
            task, project_dir, kebab_app_name, snake_module_name
        )),
        asyncio.create_task(_register_migration(
            task, project_dir, kebab_app_name, snake_module_name
        )),
        asyncio.create_task(_create_app_config(
            task, project_dir, kebab_app_name, snake_module_name,
            upper_snake_module_name
        )),
        asyncio.create_task(_append_all_enabled_env(
            task, project_dir, kebab_app_name, upper_snake_module_name
        )),
        asyncio.create_task(_append_all_disabled_env(
            task, project_dir, kebab_app_name, upper_snake_module_name
        )),
        asyncio.create_task(_append_src_template_env(
            task, project_dir, kebab_app_name, upper_snake_module_name
        )),
        asyncio.create_task(_append_deployment_template_env(
            task, project_dir, kebab_app_name, upper_snake_module_name
        )),
    )
    task.print_out('Success')


async def _create_microservice_config(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    snake_app_name: str,
    kebab_module_name: str,
    snake_module_name: str,
    upper_snake_module_name: str
):
    modules = await _create_automation_json_config(
        task, project_dir, snake_app_name, snake_module_name
    )
    return await asyncio.gather(
        asyncio.create_task(_add_docker_compose_service(
            task, modules, project_dir, kebab_app_name,
            snake_app_name, kebab_module_name, snake_module_name,
            upper_snake_module_name
        )),
        asyncio.create_task(_append_compose_env(
            task, modules, project_dir, snake_app_name, upper_snake_module_name
        ))
    )


async def _add_docker_compose_service(
    task: Task,
    modules: List[str],
    project_dir: str,
    kebab_app_name: str,
    snake_app_name: str,
    kebab_module_name: str,
    snake_module_name: str,
    upper_snake_module_name: str
):
    module_app_port = 8080 + len(modules)
    module_app_port_str = str(module_app_port)
    docker_compose_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'docker-compose.yml'
    )
    app_container_port_env_name = f'APP_{upper_snake_module_name}_MODULE_PORT'
    app_container_port_env = '${' + app_container_port_env_name + ':-' + module_app_port_str + '}' # noqa
    app_host_port_env_name = f'APP_{upper_snake_module_name}_HOST_MODULE_PORT'
    app_host_port_env = '${' + app_host_port_env_name + ':-' + module_app_port_str + '}' # noqa
    service_definition = _get_new_docker_compose_service_definition(
        kebab_app_name, snake_app_name, kebab_module_name, snake_module_name,
        upper_snake_module_name, app_host_port_env, app_container_port_env
    )
    task.print_out(f'Add service at: {docker_compose_file_path}')
    add_services(
        file_name=docker_compose_file_path,
        new_services=service_definition,
    )


def _get_new_docker_compose_service_definition(
    kebab_app_name: str,
    snake_app_name: str,
    kebab_module_name: str,
    snake_module_name: str,
    upper_snake_module_name: str,
    app_host_port_env: str,
    app_container_port_env: str
) -> Mapping[str, Any]:
    return {
        f'{kebab_app_name}-{kebab_module_name}-service': {
            'build': {
                'dockerfile': 'Dockerfile',
                'context': './src'
            },
            'image': '${IMAGE:-' + kebab_app_name + '}',
            'container_name': f'{snake_app_name}_{snake_module_name}_service',
            'hostname': f'snake_app_name_{snake_module_name}',
            'env_file': [
                'src/template.env',
                'all-module-disabled.env'
            ],
            'environment': {
                'APP_NAME': '${APP_NAME:-' + kebab_app_name + '}-' + f'{kebab_module_name}-service', # noqa
                'APP_PORT': app_container_port_env,  # noqa
                'APP_RMQ_CONNECTION': '${APP_CONTAINER_RMQ_CONNECTION:-amqp://guest:guest@rabbitmq/}', # noqa
                'APP_KAFKA_BOOTSTRAP_SERVERS': '${APP_CONTAINER_KAFKA_BOOTSTRAP_SERVERS:-redpanda:9092}', # noqa
                'APP_DB_CONNECTION': '${APP_CONTAINER_DB_CONNECTION:-postgresql+psycopg2://postgres:admin@postgresql:5432/' + snake_app_name + '}', # noqa
                'APP_ENABLE_MESSAGE_CONSUMER': 'true',
                'APP_ENABLE_RPC_SERVER': 'true',
                'APP_ENABLE_API': 'false',
                'APP_ENABLE_FRONTEND': 'false',
                f'APP_ENABLE_{upper_snake_module_name}_MODULE': 'true',
            },
            'ports': [
                f'{app_host_port_env}:{app_container_port_env}'
            ],
            'restart': 'on-failure',
            'profiles': [
                'microservices'
            ],
            'healthcheck': {
                'test': [
                    "CMD-SHELL",
                    "curl --fail http://localhost:${APP_PORT:-8080}/readiness || exit 1" # noqa
                ],
                'interval': '5s',
                'timeout': '3s',
                'retries': 10
            }
        }
    }


async def _append_compose_env(
    task: Task,
    modules: List[str],
    project_dir: str,
    snake_app_name: str,
    upper_snake_module_name: str
):
    compose_template_env_path = os.path.join(
        project_dir, '_automate', snake_app_name, 'config',
        'docker-compose.env'
    )
    compose_env_map = dotenv_values(compose_template_env_path)
    host_port = int(compose_env_map.get('APP_GATEWAY_HOST_PORT', '8080'))
    module_app_port = host_port + len(modules)
    new_env_str = '\n'.join([
        f'APP_{upper_snake_module_name}_HOST_MODULE_PORT={module_app_port}',
        f'APP_{upper_snake_module_name}_MODULE_PORT={module_app_port}',
    ])
    task.print_out(f'Add new environment to: {compose_template_env_path}')
    await append_text_file_async(compose_template_env_path, new_env_str)


async def _append_src_template_env(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    upper_snake_module_name: str
):
    src_template_env_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'template.env'
    )
    new_env_str = '\n'.join([
        f'APP_ENABLE_{upper_snake_module_name}_MODULE=true',
    ])
    task.print_out(f'Add new environment to: {src_template_env_path}')
    await append_text_file_async(src_template_env_path, new_env_str)


async def _append_deployment_template_env(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    upper_snake_module_name: str
):
    deployment_template_env_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'deployment', 'template.env'
    )
    new_env_str = '\n'.join([
        f'REPLICA_{upper_snake_module_name}_SERVICE=1',
    ])
    task.print_out(f'Add new environment to: {deployment_template_env_path}')
    await append_text_file_async(deployment_template_env_path, new_env_str)


async def _append_all_enabled_env(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    upper_snake_module_name: str
):
    all_enabled_env_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'all-module-enabled.env'
    )
    task.print_out(f'Add new environment to: {all_enabled_env_path}')
    await append_text_file_async(
        all_enabled_env_path,
        f'APP_ENABLE_{upper_snake_module_name}_MODULE=true'
    )


async def _append_all_disabled_env(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    upper_snake_module_name: str
):
    all_disabled_env_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'all-module-disabled.env'
    )
    task.print_out(f'Add new environment to: {all_disabled_env_path}')
    await append_text_file_async(
        all_disabled_env_path,
        f'APP_ENABLE_{upper_snake_module_name}_MODULE=false'
    )


async def _create_app_config(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    snake_module_name: str,
    upper_snake_module_name: str
):
    config_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'config.py'
    )
    config_code = '\n'.join([
        f'app_enable_{snake_module_name}_module = str_to_boolean(os.environ.get(',  # noqa
        f"    'APP_ENABLE_{upper_snake_module_name}_MODULE', 'true'"
        '))'
    ])
    task.print_out(f'Read config from: {config_file_path}')
    code = await read_text_file_async(config_file_path)
    code += '\n' + config_code
    task.print_out(f'Write config to: {config_file_path}')
    await write_text_file_async(config_file_path, code)


async def _create_automation_json_config(
    task: Task,
    project_dir: str,
    snake_app_name: str,
    snake_module_name: str,
):
    json_modules_file_path = os.path.join(
        project_dir, '_automate', snake_app_name, 'config', 'modules.json'
    )
    task.print_out(f'Read json config from: {json_modules_file_path}')
    json_str = await read_text_file_async(json_modules_file_path)
    task.print_out(f'Add "{snake_module_name}" to json config')
    modules: List[str] = jsons.loads(json_str)
    modules.append(snake_module_name)
    json_str = jsons.dumps(modules)
    task.print_out(f'Write new json config to: {json_modules_file_path}')
    await write_text_file_async(json_modules_file_path, json_str)
    return modules


async def _register_migration(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    snake_module_name: str
):
    app_migration_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'migrate.py'
    )
    import_module_path = '.'.join([
        'module', snake_module_name, 'migrate'
    ])
    function_name = f'migrate_{snake_module_name}'
    task.print_out(f'Read code from: {app_migration_file_path}')
    code = await read_text_file_async(app_migration_file_path)
    task.print_out(
        f'Add import "{function_name}" from "{import_module_path}" to the code'
    )
    code = add_import_module(
        code=code,
        module_path=import_module_path,
        resource=function_name
    )
    task.print_out(f'Add "{function_name}" call to the code')
    code = append_code_to_function(
        code=code,
        function_name='migrate',
        new_code=f'await {function_name}()'
    )
    task.print_out(f'Write modified code to: {app_migration_file_path}')
    await write_text_file_async(app_migration_file_path, code)


async def _register_module(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    snake_module_name: str
):
    app_main_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'main.py'
    )
    import_module_path = '.'.join([
        'module', snake_module_name, 'register_module'
    ])
    function_name = f'register_{snake_module_name}'
    task.print_out(f'Read code from: {app_main_file_path}')
    code = await read_text_file_async(app_main_file_path)
    task.print_out(
        f'Add import "{function_name}" from "{import_module_path}" to the code'
    )
    code = add_import_module(
        code=code,
        module_path=import_module_path,
        resource=function_name
    )
    task.print_out(f'Add "{function_name}" call to the code')
    code = add_function_call(
        code=code,
        function_name=function_name,
        parameters=[]
    )
    task.print_out(f'Write modified code to: {app_main_file_path}')
    await write_text_file_async(app_main_file_path, code)



from pulumi_kubernetes.helm.v3 import Chart, LocalChartOpts
from _common import (
    NAMESPACE, POSTGRESQL_AUTH_PASSWORD, POSTGRESQL_AUTH_USERNAME,
    POSTGRESQL_AUTH_POSTGRES_PASSWORD, POSTGRESQL_DB
)


def create_postgresql() -> Chart:
    postgresql = Chart(
        'postgresql',
        LocalChartOpts(
            path='./helm-charts/postgresql',
            namespace=NAMESPACE,
            values={
                'auth': {
                    'postgresPassword': POSTGRESQL_AUTH_POSTGRES_PASSWORD,
                    'username': POSTGRESQL_AUTH_USERNAME,
                    'password': POSTGRESQL_AUTH_PASSWORD,
                    'database': POSTGRESQL_DB,
                },
                'resources': {
                    'limits': {
                        'cpu': '250m',
                        'memory': '256Mi'
                    },
                    'requests': {
                        'cpu': '250m',
                        'memory': '156Mi'
                    },
                }
            },
            skip_await=True
        ),
    )
    return postgresql



from typing import Iterable, Optional
from termcolor import colored as term_colored, COLORS
import random


def get_random_color():
    colors = [
        'green', 'yellow', 'blue', 'magenta', 'cyan',
        'light_green', 'light_yellow', 'light_blue',
        'light_magenta', 'light_cyan'
    ]
    return random.choice(colors)


def is_valid_color(color: str) -> bool:
    return color in COLORS


def colored(
    text: str,
    color: Optional[str] = None,
    on_color: Optional[str] = None,
    attrs: Optional[Iterable[str]] = None
) -> str:
    return term_colored(text, color, on_color, attrs)



from zrb import DockerComposeTask, runner
from zrb.builtin._group import project_group
from ._common import (
    RESOURCE_DIR, SKIP_CONTAINER_EXECUTION,
    app_container_checker, rabbitmq_checker, rabbitmq_management_checker,
    redpanda_console_checker, kafka_outside_checker,
    kafka_plaintext_checker, pandaproxy_outside_checker,
    pandaproxy_plaintext_checker,
    local_input, mode_input, host_input, https_input, image_input,
    compose_env_file, image_env
)
from .image import build_snake_app_name_image

compose_env_prefix = 'CONTAINER_ENV_PREFIX'
all_compose_profiles = 'monolith,microservices,kafka,rabbitmq'
start_broker_compose_profile = '{{env.get("APP_BROKER_TYPE", "rabbitmq")}}'
start_mode_compose_profile = '{{input.get("snake_app_name_mode", "monolith")}}'
start_compose_profiles = ','.join([
    start_broker_compose_profile, start_mode_compose_profile
])

###############################################################################
# Task Definitions
###############################################################################

remove_snake_app_name_container = DockerComposeTask(
    icon='💨',
    name='remove-kebab-app-name-container',
    description='Rumove human readable app name container',
    group=project_group,
    cwd=RESOURCE_DIR,
    setup_cmd=f'export COMPOSE_PROFILES={all_compose_profiles}',
    compose_cmd='down',
    compose_env_prefix=compose_env_prefix,
    env_files=[compose_env_file],
    envs=[
        image_env,
    ],
)
runner.register(remove_snake_app_name_container)

start_snake_app_name_container = DockerComposeTask(
    icon='🐳',
    name='start-kebab-app-name-container',
    description='Start human readable app name container',
    group=project_group,
    inputs=[
        local_input,
        mode_input,
        host_input,
        https_input,
        image_input,
    ],
    skip_execution=SKIP_CONTAINER_EXECUTION,
    upstreams=[
        build_snake_app_name_image,
        remove_snake_app_name_container
    ],
    cwd=RESOURCE_DIR,
    setup_cmd=f'export COMPOSE_PROFILES={start_compose_profiles}',
    compose_cmd='up',
    compose_env_prefix=compose_env_prefix,
    env_files=[compose_env_file],
    envs=[
        image_env,
    ],
    checkers=[
        app_container_checker,
        rabbitmq_checker,
        rabbitmq_management_checker,
        kafka_outside_checker,
        kafka_plaintext_checker,
        redpanda_console_checker,
        pandaproxy_outside_checker,
        pandaproxy_plaintext_checker,
    ]
)
runner.register(start_snake_app_name_container)

stop_snake_app_name_container = DockerComposeTask(
    icon='⛔',
    name='stop-kebab-app-name-container',
    description='Stop human readable app name container',
    group=project_group,
    cwd=RESOURCE_DIR,
    setup_cmd=f'export COMPOSE_PROFILES={all_compose_profiles}',
    compose_cmd='stop',
    compose_env_prefix=compose_env_prefix,
    env_files=[compose_env_file],
    envs=[
        image_env,
    ],
)
runner.register(stop_snake_app_name_container)



from typing import Any, Optional
from core.messagebus.messagebus import (
    Publisher, MessageSerializer, must_get_message_serializer
)
from core.messagebus.kafka.admin import (
    KafkaAdmin, must_get_kafka_admin
)
from aiokafka import AIOKafkaProducer
from aiokafka.producer.producer import _missing, DefaultPartitioner
from pydantic import BaseModel

import logging
import asyncio


class KafkaPublisher(Publisher):
    def __init__(
        self,
        logger: logging.Logger,
        bootstrap_servers: str = 'localhost',
        client_id: Optional[Any] = None,
        metadata_max_age_ms=300000,
        request_timeout_ms=40000,
        api_version='auto',
        acks=_missing,
        key_serializer=None,
        value_serializer=None,
        compression_type=None,
        max_batch_size=16384,
        partitioner=DefaultPartitioner(),
        max_request_size=1048576,
        linger_ms=0,
        send_backoff_ms=100,
        retry_backoff_ms=100,
        security_protocol="PLAINTEXT",
        ssl_context=None,
        connections_max_idle_ms=540000,
        enable_idempotence=False,
        transactional_id=None,
        transaction_timeout_ms=60000,
        sasl_mechanism="PLAIN",
        sasl_plain_password=None,
        sasl_plain_username=None,
        sasl_kerberos_service_name='kafka',
        sasl_kerberos_domain_name=None,
        sasl_oauth_token_provider=None,
        serializer: Optional[MessageSerializer] = None,
        kafka_admin: Optional[KafkaAdmin] = None,
        retry: int = 3,
        retry_interval: int = 3,
        identifier='kafka-publisher'
    ):
        self.logger = logger
        self.serializer = must_get_message_serializer(serializer)
        self.kafka_admin = must_get_kafka_admin(
            logger=logger,
            kafka_admin=kafka_admin,
            bootstrap_servers=bootstrap_servers,
            security_protocol=security_protocol,
            sasl_mechanism=sasl_mechanism,
            sasl_plain_password=sasl_plain_password,
            sasl_plain_username=sasl_plain_username,
            sasl_kerberos_service_name=sasl_kerberos_service_name,
            sasl_kerberos_domain_name=sasl_kerberos_domain_name,
            sasl_oauth_token_provider=sasl_oauth_token_provider,
        )
        self.producer: Optional[AIOKafkaProducer] = None
        self.bootstrap_servers = bootstrap_servers
        self.client_id = client_id
        self.metadata_max_age_ms = metadata_max_age_ms
        self.request_timeout_ms = request_timeout_ms
        self.api_version = api_version
        self.acks = acks
        self.key_serializer = key_serializer
        self.value_serializer = value_serializer
        self.compression_type = compression_type
        self.max_batch_size = max_batch_size
        self.partitioner = partitioner
        self.max_request_size = max_request_size
        self.linger_ms = linger_ms
        self.send_backoff_ms = send_backoff_ms
        self.retry_backoff_ms = retry_backoff_ms
        self.security_protocol = security_protocol
        self.ssl_context = ssl_context
        self.connections_max_idle_ms = connections_max_idle_ms
        self.enable_idempotence = enable_idempotence
        self.transactional_id = transactional_id
        self.transaction_timeout_ms = transaction_timeout_ms
        self.sasl_mechanism = sasl_mechanism
        self.sasl_plain_password = sasl_plain_password
        self.sasl_plain_username = sasl_plain_username
        self.sasl_kerberos_service_name = sasl_kerberos_service_name
        self.sasl_kerberos_domain_name = sasl_kerberos_domain_name
        self.sasl_oauth_token_provider = sasl_oauth_token_provider
        self.retry = retry
        self.retry_interval = retry_interval
        self.identifier = identifier

    async def publish(self, event_name: str, message: Any):
        await self.kafka_admin.create_events([event_name])
        topic_name = self.kafka_admin.get_topic_name(event_name)
        if isinstance(message, BaseModel):
            message = message.dict()
        for attempt in range(self.retry):
            try:
                await self._connect()
                encoded_value = self.serializer.encode(event_name, message)
                self.logger.info(
                    f'🐼 [{self.identifier}] Publish to "{topic_name}": ' +
                    f'{message}'
                )
                return await self.producer.send_and_wait(
                    topic_name, encoded_value
                )
            except (asyncio.CancelledError, GeneratorExit, Exception) as e:
                self.logger.error(
                    f'🐼 [{self.identifier}] Failed to publish message: {e}'
                )
                await self._disconnect()
                await asyncio.sleep(self.retry_interval)
                continue
        self.logger.error(
            f'🐼 [{self.identifier}] Failed to publish message after ' +
            f'{self.retry} attempts'
        )
        raise RuntimeError('Failed to publish message after retrying')

    async def _connect(self):
        self.logger.info(f'🐼 [{self.identifier}] Create kafka producer')
        self.producer = AIOKafkaProducer(
            bootstrap_servers=self.bootstrap_servers,
            client_id=self.client_id,
            metadata_max_age_ms=self.metadata_max_age_ms,
            request_timeout_ms=self.request_timeout_ms,
            api_version=self.api_version,
            acks=self.acks,
            key_serializer=self.key_serializer,
            value_serializer=self.value_serializer,
            compression_type=self.compression_type,
            max_batch_size=self.max_batch_size,
            partitioner=self.partitioner,
            max_request_size=self.max_request_size,
            linger_ms=self.linger_ms,
            send_backoff_ms=self.send_backoff_ms,
            retry_backoff_ms=self.retry_backoff_ms,
            security_protocol=self.security_protocol,
            ssl_context=self.ssl_context,
            connections_max_idle_ms=self.connections_max_idle_ms,
            enable_idempotence=self.enable_idempotence,
            transactional_id=self.transactional_id,
            transaction_timeout_ms=self.transaction_timeout_ms,
            sasl_mechanism=self.sasl_mechanism,
            sasl_plain_password=self.sasl_plain_password,
            sasl_plain_username=self.sasl_plain_username,
            sasl_kerberos_service_name=self.sasl_kerberos_service_name,
            sasl_kerberos_domain_name=self.sasl_kerberos_domain_name,
            sasl_oauth_token_provider=self.sasl_oauth_token_provider,
        )
        self.logger.info(f'🐼 [{self.identifier}] Start kafka producer')
        await self.producer.start()
        self.logger.info(f'🐼 [{self.identifier}] Kafka producer started')
        return self

    async def _disconnect(self):
        self.logger.info(f'🐼 [{self.identifier}] Stop kafka producer')
        if self.producer is not None:
            await self.producer.stop()
        self.logger.info(f'🐼 [{self.identifier}] Kafka producer stopped')
        self.producer = None



from typing import Any
from ..._group import project_add_group
from ....task.task import Task
from ....task.decorator import python_task
from ....task.resource_maker import ResourceMaker
from ....runner import runner
from .._common.input import (
    project_dir_input, app_name_input, app_image_name_input, http_port_input,
    env_prefix_input
)
from .._common.helper import validate_project_dir, create_register_app_module
from .._common.lock import new_task_app_lock
from ..project_task.task_factory import (
    create_add_project_automation_task, create_register_app_task
)
from ....helper import util

import os

# Common definitions

current_dir = os.path.dirname(__file__)


@python_task(
    name='validate',
    inputs=[project_dir_input, app_name_input],
)
async def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir')
    validate_project_dir(project_dir)
    app_name = kwargs.get('app_name')
    automation_dir = os.path.join(
        project_dir, '_automate', util.to_snake_case(app_name)
    )
    if os.path.exists(automation_dir):
        raise Exception(
            f'Automation directory already exists: {automation_dir}'
        )
    source_dir = os.path.join(
        project_dir, 'src', f'{util.to_kebab_case(app_name)}'
    )
    if os.path.exists(source_dir):
        raise Exception(f'Source already exists: {source_dir}')


copy_resource = ResourceMaker(
    name='copy-resource',
    inputs=[
        project_dir_input,
        app_name_input,
        app_image_name_input,
        http_port_input,
        env_prefix_input,
    ],
    upstreams=[validate],
    replacements={
        'appName': '{{input.app_name}}',
        'httpPort': '{{util.coalesce(input.http_port, "3000")}}',
        'ENV_PREFIX': '{{util.coalesce(input.env_prefix, "MY").upper()}}',
        'app-image-name': '{{input.app_image_name}}'
    },
    template_path=os.path.join(current_dir, 'template'),
    destination_path='{{ input.project_dir }}',
    locks=[new_task_app_lock],
    excludes=['*/deployment/venv']
)

register_local_app_module = create_register_app_module(
    module='local', upstreams=[copy_resource]
)

register_container_app_module = create_register_app_module(
    module='container', upstreams=[register_local_app_module]
)

register_image_app_module = create_register_app_module(
    module='image', upstreams=[register_container_app_module]
)

register_deployment_app_module = create_register_app_module(
    module='deployment', upstreams=[register_image_app_module]
)

add_project_task = create_add_project_automation_task(
    upstreams=[copy_resource]
)

register_start = create_register_app_task(
    task_name='register-start',
    project_automation_file_name='start_project.py',
    project_automation_task_name='start',
    app_automation_file_name='local.py',
    app_automation_task_var_name_tpl='start_{snake_app_name}',
    upstreams=[add_project_task]
)

register_start_container = create_register_app_task(
    task_name='register-start-container',
    project_automation_file_name='start_project_containers.py',
    project_automation_task_name='start-containers',
    app_automation_file_name='container.py',
    app_automation_task_var_name_tpl='start_{snake_app_name}_container',
    upstreams=[add_project_task]
)

register_stop_container = create_register_app_task(
    task_name='register-stop-container',
    project_automation_file_name='stop_project_containers.py',
    project_automation_task_name='stop-containers',
    app_automation_file_name='container.py',
    app_automation_task_var_name_tpl='stop_{snake_app_name}_container',
    upstreams=[add_project_task]
)

register_remove_container = create_register_app_task(
    task_name='register-remove-container',
    project_automation_file_name='remove_project_containers.py',
    project_automation_task_name='remove-containers',
    app_automation_file_name='container.py',
    app_automation_task_var_name_tpl='remove_{snake_app_name}_container',
    upstreams=[add_project_task]
)

register_push_image = create_register_app_task(
    task_name='register-push-image',
    project_automation_file_name='push_project_images.py',
    project_automation_task_name='push-images',
    app_automation_file_name='image.py',
    app_automation_task_var_name_tpl='push_{snake_app_name}_image',
    upstreams=[add_project_task]
)

register_build_image = create_register_app_task(
    task_name='register-build-image',
    project_automation_file_name='build_project_images.py',
    project_automation_task_name='build-images',
    app_automation_file_name='image.py',
    app_automation_task_var_name_tpl='build_{snake_app_name}_image',
    upstreams=[add_project_task]
)

register_deploy = create_register_app_task(
    task_name='register-deploy',
    project_automation_file_name='deploy_project.py',
    project_automation_task_name='deploy',
    app_automation_file_name='deployment.py',
    app_automation_task_var_name_tpl='deploy_{snake_app_name}',
    upstreams=[add_project_task]
)

register_destroy = create_register_app_task(
    task_name='register-destroy',
    project_automation_file_name='destroy_project.py',
    project_automation_task_name='destroy',
    app_automation_file_name='deployment.py',
    app_automation_task_var_name_tpl='destroy_{snake_app_name}',
    upstreams=[add_project_task]
)


@python_task(
    name='simple-python-app',
    group=project_add_group,
    upstreams=[
        register_local_app_module,
        register_container_app_module,
        register_image_app_module,
        register_deployment_app_module,
        register_start,
        register_start_container,
        register_stop_container,
        register_remove_container,
        register_build_image,
        register_push_image,
        register_deploy,
        register_destroy
    ],
    runner=runner
)
async def add_simple_python_app(*args: Any, **kwargs: Any):
    task: Task = kwargs.get('_task')
    task.print_out('Success')



from zrb import CmdTask, Env, EnvFile, PathChecker, runner
from zrb.builtin._group import project_group
from ._common import (
    APP_FRONTEND_DIR, APP_FRONTEND_BUILD_DIR, CURRENT_DIR,
    APP_TEMPLATE_ENV_FILE_NAME, local_app_port_env
)
import os

###############################################################################
# Task Definitions
###############################################################################

build_snake_app_name_frontend = CmdTask(
    icon='🚤',
    name='build-kebab-app-name-frontend',
    description='Build frontend for human readable app name',
    group=project_group,
    cwd=APP_FRONTEND_DIR,
    cmd_path=os.path.join(CURRENT_DIR, 'cmd', 'build-frontend.sh'),
    checkers=[
        PathChecker(
            name='check-kebab-app-name-frontend-build',
            path=APP_FRONTEND_BUILD_DIR
        )
    ],
    env_files=[
        EnvFile(
            env_file=APP_TEMPLATE_ENV_FILE_NAME, prefix='ENV_PREFIX'
        ),
    ],
    envs=[
        local_app_port_env,
        Env(name='WATCH', os_name='', default='1')
    ]
)
runner.register(build_snake_app_name_frontend)

build_snake_app_name_frontend_once = CmdTask(
    icon='🚤',
    name='build-kebab-app-name-frontend_once',
    description='Build frontend for human readable app name',
    group=project_group,
    cwd=APP_FRONTEND_DIR,
    cmd_path=os.path.join(CURRENT_DIR, 'cmd', 'build-frontend.sh'),
    checkers=[
        PathChecker(
            name='check-kebab-app-name-frontend-build',
            path=APP_FRONTEND_BUILD_DIR
        )
    ],
    env_files=[
        EnvFile(
            env_file=APP_TEMPLATE_ENV_FILE_NAME, prefix='ENV_PREFIX'
        ),
    ],
    envs=[
        local_app_port_env,
        Env(name='WATCH', os_name='', default='0')
    ]
)



from asgi_lifespan import LifespanManager
from httpx import AsyncClient
from src.main import app
import pytest


@pytest.fixture()
async def test_client_generator(scope='module'):
    async with LifespanManager(app):
        async with AsyncClient(app=app, base_url="http://localhost") as ac:
            yield ac



from zrb import CmdTask, DockerComposeTask, Task, Env, EnvFile, runner
from zrb.builtin._group import project_group
from ._common import (
    CURRENT_DIR, APP_DIR, RESOURCE_DIR, APP_TEMPLATE_ENV_FILE_NAME,
    SKIP_SUPPORT_CONTAINER_EXECUTION, SKIP_LOCAL_MONOLITH_EXECUTION,
    SKIP_LOCAL_MICROSERVICES_EXECUTION,
    rabbitmq_checker, rabbitmq_management_checker,
    redpanda_console_checker, kafka_outside_checker,
    kafka_plaintext_checker, pandaproxy_outside_checker,
    pandaproxy_plaintext_checker, app_local_checker,
    local_input, mode_input, host_input, https_input, image_input,
    local_app_port_env, local_app_broker_type_env,
)
from .image import build_snake_app_name_image
from .frontend import build_snake_app_name_frontend
from .container import remove_snake_app_name_container
from .local_microservices import get_start_microservices
import os

support_compose_env_prefix = 'CONTAINER_ENV_PREFIX'
start_broker_compose_profile = '{{env.get("APP_BROKER_TYPE", "rabbitmq")}}'

###############################################################################
# Task Definitions
###############################################################################

start_snake_app_name_support_container = DockerComposeTask(
    icon='🐳',
    name='start-kebab-app-name-support-container',
    description='Start human readable app name container',
    inputs=[
        local_input,
        host_input,
        https_input,
        image_input,
    ],
    skip_execution=SKIP_SUPPORT_CONTAINER_EXECUTION,
    upstreams=[
        build_snake_app_name_image,
        remove_snake_app_name_container
    ],
    cwd=RESOURCE_DIR,
    setup_cmd=f'export COMPOSE_PROFILES={start_broker_compose_profile}',
    compose_cmd='up',
    compose_env_prefix=support_compose_env_prefix,
    envs=[
        local_app_broker_type_env,
        local_app_port_env,
    ],
    checkers=[
        rabbitmq_checker,
        rabbitmq_management_checker,
        kafka_outside_checker,
        kafka_plaintext_checker,
        redpanda_console_checker,
        pandaproxy_outside_checker,
        pandaproxy_plaintext_checker,
    ]
)

prepare_snake_app_name_backend = CmdTask(
    icon='🚤',
    name='prepare-kebab-app-name-backend',
    description='Prepare backend for human readable app name',
    group=project_group,
    cwd=APP_DIR,
    cmd_path=os.path.join(CURRENT_DIR, 'cmd', 'prepare-backend.sh'),
)
runner.register(prepare_snake_app_name_backend)

start_monolith_snake_app_name = CmdTask(
    icon='🚤',
    name='start-monolith-kebab-app-name',
    inputs=[
        local_input,
        mode_input,
        host_input,
        https_input
    ],
    skip_execution=SKIP_LOCAL_MONOLITH_EXECUTION,
    upstreams=[
        start_snake_app_name_support_container,
        build_snake_app_name_frontend,
        prepare_snake_app_name_backend,
    ],
    cwd=APP_DIR,
    env_files=[
        EnvFile(
            env_file=APP_TEMPLATE_ENV_FILE_NAME, prefix='ENV_PREFIX'
        ),
    ],
    envs=[
        local_app_broker_type_env,
        local_app_port_env,
    ],
    cmd_path=os.path.join(CURRENT_DIR, 'cmd', 'start.sh'),
    checkers=[
        app_local_checker,
    ]
)

start_snake_app_name_gateway = CmdTask(
    icon='🚪',
    name='start-kebab-app-name-gateway',
    inputs=[
        local_input,
        mode_input,
        host_input,
        https_input
    ],
    skip_execution=SKIP_LOCAL_MICROSERVICES_EXECUTION,
    upstreams=[
        start_snake_app_name_support_container,
        build_snake_app_name_frontend,
        prepare_snake_app_name_backend,
    ],
    cwd=APP_DIR,
    env_files=[
       EnvFile(
            env_file=APP_TEMPLATE_ENV_FILE_NAME, prefix='ENV_PREFIX'
        ),
    ],
    envs=[
        local_app_broker_type_env,
        local_app_port_env,
        Env(name='APP_DB_AUTO_MIGRATE', default='false', os_name=''),
        Env(name='APP_ENABLE_MESSAGE_CONSUMER', default='false', os_name=''),
        Env(name='APP_ENABLE_RPC_SERVER', default='false', os_name=''),
    ],
    cmd_path=os.path.join(CURRENT_DIR, 'cmd', 'start.sh'),
    checkers=[
        app_local_checker,
    ]
)
runner.register(start_snake_app_name_gateway)

start_microservices = get_start_microservices(
    upstreams=[
        start_snake_app_name_support_container,
        build_snake_app_name_frontend,
        prepare_snake_app_name_backend,
    ]
)

start_snake_app_name = Task(
    icon='🚤',
    name='start-kebab-app-name',
    description='Start human readable app name',
    group=project_group,
    upstreams=[
        start_monolith_snake_app_name,
        start_snake_app_name_gateway,
    ] + start_microservices,
    run=lambda *args, **kwargs: kwargs.get('_task').print_out('👌')
)
runner.register(start_snake_app_name)



from zrb import CmdTask, runner
from zrb.builtin._group import project_group

###############################################################################
# Task Definitions
###############################################################################

snake_task_name = CmdTask(
    name='kebab-task-name',
    description='human readable task name',
    group=project_group,
    cmd=[
        'echo human readable task name'
    ]
)
runner.register(snake_task_name)



from typing import List
from zrb import CmdTask, Env, HTTPChecker, Task, EnvFile, runner
from zrb.helper.util import to_snake_case, to_kebab_case
from ._common import (
    CURRENT_DIR, APP_DIR, SKIP_LOCAL_MICROSERVICES_EXECUTION,
    APP_TEMPLATE_ENV_FILE_NAME, MODULES,
    local_app_broker_type_env, local_input, mode_input, host_input, https_input
)
import os


def get_start_microservices(upstreams: List[Task]) -> List[Task]:
    service_disable_module_envs: List[Env] = []
    for index, module in enumerate(MODULES):
        snake_module_name = to_snake_case(module)
        upper_snake_module_name = snake_module_name.upper()
        service_disable_module_envs.append(Env(
            name=f'APP_ENABLE_{upper_snake_module_name}_MODULE',
            os_name='',
            default='false'
        ))
    start_microservices: List[Task] = []
    for index, module in enumerate(MODULES):
        kebab_module_name = to_kebab_case(module)
        snake_module_name = to_snake_case(module)
        upper_snake_module_name = snake_module_name.upper()
        # Create service_env_file
        service_template_env_file = EnvFile(
            env_file=APP_TEMPLATE_ENV_FILE_NAME,
            prefix=f'ENV_PREFIX_{upper_snake_module_name}'
        )
        service_env_files = [service_template_env_file]
        # Create service_envs
        service_port_str = str(httpPort + index + 1)
        service_app_port_env = Env(
            name='APP_PORT',
            os_name=f'ENV_PREFIX_{upper_snake_module_name}_APP_PORT',
            default=service_port_str
        )
        service_enable_module_env = Env(
            name=f'APP_ENABLE_{upper_snake_module_name}_MODULE',
            os_name='',
            default='true'
        )
        service_envs = service_disable_module_envs + [
            local_app_broker_type_env,
            service_app_port_env,
            service_enable_module_env
        ]
        # Create service_checker
        service_checker = HTTPChecker(
            name=f'check-kebab-app-name-{kebab_module_name}-service',
            host='{{input.snake_app_name_host}}',
            url='/readiness',
            port='{{env.APP_PORT}}',
            is_https='{{input.snake_app_name_https}}',
            skip_execution=SKIP_LOCAL_MICROSERVICES_EXECUTION
        )
        # Create start_service
        start_service = CmdTask(
            name=f'start-kebab-app-name-{kebab_module_name}-service',
            inputs=[
                local_input,
                mode_input,
                host_input,
                https_input
            ],
            skip_execution=SKIP_LOCAL_MICROSERVICES_EXECUTION,
            upstreams=upstreams,
            cwd=APP_DIR,
            env_files=service_env_files,
            envs=service_envs,
            cmd_path=os.path.join(CURRENT_DIR, 'cmd', 'start.sh'),
            checkers=[
                service_checker,
            ]
        )
        runner.register(start_service)
        start_microservices.append(start_service)
    return start_microservices
